//licence: GPL
//autor: Clemens Zeidler czeidler@gmx.de 
//basicly an incomplete port of the linux speedstep_ich driver
//also based on the speedstep centrino driver from Gergely Rózsahegyi

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <PCI.h>
#include "SpeedStep.h"

#define DEBUG 1
#if DEBUG > 0
#define ddprintf(a)    dprintf a
#else
#define    ddprintf(a)
#endif

#define MSR_IA32_MISC_ENABLE 0x1a0
#define MSR_IA32_PERF_STATUS 0x198
#define MSR_IA32_PERF_CTL	 0x199

#define PCI_VENDOR_ID_INTEL             0x8086
#define PCI_DEVICE_ID_INTEL_82801DB_12  0x24cc   /* 4-M */
#define PCI_DEVICE_ID_INTEL_82801CA_12  0x248c 	 /* 3-M */
#define PCI_DEVICE_ID_INTEL_82801BA_10	0x244c	 /* 2-M */

#define B_CPU_INTEL_PENTIUM_III_MODEL_B 0x106b //Intel P3 Tualatin

static uint32 pmbase;
static pci_module_info * pci = NULL; 
static uint32 processorType;
static cpu_state cpuState;
static pci_info info;

/*********************************************************************
 *                   GET PROCESSOR CORE SPEED IN KHZ                 *
 *********************************************************************/

static uint32 pentium3_get_frequency (unsigned int processor)
{
        /* See table 14 of p3_ds.pdf and table 22 of 29834003.pdf */
	struct {
		unsigned int ratio;	/* Frequency Multiplier (x10) */
		uint8 bitmap;		/* power on configuration bits
					[27, 25:22] (in MSR 0x2a) */
	} msr_decode_mult [] = {
		{ 30, 0x01 },
		{ 35, 0x05 },
		{ 40, 0x02 },
		{ 45, 0x06 },
		{ 50, 0x00 },
		{ 55, 0x04 },
		{ 60, 0x0b },
		{ 65, 0x0f },
		{ 70, 0x09 },
		{ 75, 0x0d },
		{ 80, 0x0a },
		{ 85, 0x26 },
		{ 90, 0x20 },
		{ 100, 0x2b },
		{ 0, 0xff }     /* error or unknown value */
	};

	/* PIII(-M) FSB settings: see table b1-b of 24547206.pdf */
	struct {
		unsigned int value;	/* Front Side Bus speed in MHz */
		uint8 bitmap;		/* power on configuration bits [18: 19]
					(in MSR 0x2a) */
	} msr_decode_fsb [] = {
		{  66, 0x0 },
		{ 100, 0x2 },
		{ 133, 0x1 },
		{   0, 0xff}
	};

	uint32 msr_lo, msr_tmp;
	int i = 0, j = 0;

	/* read MSR 0x2a - we only need the low 32 bits */
	rdmsr(MSR_IA32_EBL_CR_POWERON, msr_lo, msr_tmp);
	ddprintf(("SPEEDSTEP P3 - MSR_IA32_EBL_CR_POWERON: 0x%x 0x%x\n", msr_lo, msr_tmp));
	msr_tmp = msr_lo;

	/* decode the FSB */
	msr_tmp &= 0x00c0000;
	msr_tmp >>= 18;
	while (msr_tmp != msr_decode_fsb[i].bitmap) {
		if (msr_decode_fsb[i].bitmap == 0xff)
			return 0;
		i++;
	}

	/* decode the multiplier */
	if (processor == SPEEDSTEP_PROCESSOR_PIII_C_EARLY) {
		ddprintf(("SPEEDSTEP workaround for early PIIIs\n"));
		msr_lo &= 0x03c00000;
	} else
		msr_lo &= 0x0bc00000;
	msr_lo >>= 22;
	while (msr_lo != msr_decode_mult[j].bitmap) {
		if (msr_decode_mult[j].bitmap == 0xff)
			return 0;
		j++;
	}

	ddprintf(("SPEEDSTEP speed is %u\n", (msr_decode_mult[j].ratio * msr_decode_fsb[i].value * 100)));

	return (msr_decode_mult[j].ratio * msr_decode_fsb[i].value * 100);
}


/**
 * speedstep_set_state - set the SpeedStep state
 * @state: new processor frequency state (SPEEDSTEP_LOW or SPEEDSTEP_HIGH)
 *
 *   Tries to change the SpeedStep state.
 */
static void speedstep_set_state (unsigned int state)
{
	uint8 pm2_blk;
	uint8 value;
	cpu_status cp;
	
	if (state > 0x1)
		return;

	/* Disable IRQs */
	cp = disable_interrupts(); 

	/* read state */
	value = pci->read_io_8(pmbase + 0x50);

	ddprintf(("SPEEDSTEP read at pmbase 0x%x + 0x50 returned 0x%x\n", pmbase, value));

	/* write new state */
	value &= 0xFE;
	value |= state;

	ddprintf(("SPEEDSTEP writing 0x%x to pmbase 0x%x + 0x50\n", value, pmbase));

	/* Disable bus master arbitration */
	pm2_blk = read_io_8(pmbase + 0x20);
	pm2_blk |= 0x01;
	write_io_8((pmbase + 0x20), pm2_blk);

	/* Actual transition */
	write_io_8((pmbase + 0x50), value);

	/* Restore bus master arbitration */
	pm2_blk &= 0xfe;
	write_io_8((pmbase + 0x20), pm2_blk);

	/* check if transition was successful */
	value = read_io_8(pmbase + 0x50);

	/* Enable IRQs */
	restore_interrupts(cp);

	ddprintf(("SPEEDSTEP read at pmbase 0x%x + 0x50 returned 0x%x\n", pmbase, value));

	if (state == (value & 0x1)) {
		cpuState.state = state;
		cpuState.frequency = pentium3_get_frequency(processorType);
		ddprintf(("SPEEDSTEP change to %u MHz succeeded\n", (cpuState.frequency / 1000)));
		//ddprintf(("SPEEDSTEP changed freq\n"));
	} else {
		//printk (KERN_ERR "cpufreq: change failed - I/O error\n");
		ddprintf(("SPEEDSTEP KERN_ERR cpufreq: change failed - I/O error\n"));
	}

	return;
}


/**
 * speedstep_find_register - read the PMBASE address
 *
 * Returns: -ENODEV if no register could be found
 */
static int speedstep_find_register (pci_info *info)
{
	/* get PMBASE */
	pmbase = pci->read_pci_config(info->bus, info->device, info->function, 0x40, 4);
	if (!(pmbase & 0x01)) {
		ddprintf(("SPEEDSTEP could not find speedstep register 0x%x\n", pmbase));
		return ENODEV;
	}
	pmbase &= 0xFFFFFFFE;
	if (!pmbase) {
		ddprintf(("SPEEDSTEP could not find speedstep register 2 0x%x\n", pmbase));
		return ENODEV;
	}

	ddprintf(("SPEEDSTEP pmbase is 0x%x\n", pmbase));
	return B_OK;
}


/**
 * speedstep_activate - activate SpeedStep control in the chipset
 *
 *   Tries to activate the SpeedStep status and control registers.
 * Returns -EINVAL on an unsupported chipset, and zero on success.
 */
static int speedstep_activate (pci_info *info)
{
	uint16 value = 0;
	
	value = pci->read_pci_config(info->bus, info->device, info->function, 0x00A0, 2);
						
	if (!(value & 0x08)) {
		value |= 0x08;
		ddprintf(("SPEEDSTEP activating SpeedStep (TM) registers\n"));
		pci->write_pci_config(info->bus, info->device, info->function, 0x00A0, 2,  value);
	}

	return B_OK;
}


// ----------------------------------------------------------------------------------------------
status_t init_hardware (void)
{
	uint32 ebx, msr_lo, msr_hi;
	system_info sysInfo;
	int relaxed_check = 0;
		
	get_system_info(&sysInfo);
	switch(sysInfo.cpu_type){
		case B_CPU_INTEL_PENTIUM_III_MODEL_B:
			//ebx is 0x04 for desktop p3 and 0x06 for mobile p3
			ebx = cpuid_ebx(0x00000001);
			ebx&= 0x000000ff;
			ddprintf(("SPEEDSTEP: init_hardware(): cpuid_ebx(0x00000001)& 0x000000ff = 0x%04x\n", ebx));
			if(ebx != 0x06){
				return B_ERROR;
			}
			processorType = SPEEDSTEP_PROCESSOR_PIII_T;
        	//return B_OK;
        	break;
        case B_CPU_INTEL_PENTIUM_III_MODEL_8:
        	/* all mobile PIII Coppermines have FSB 100 MHz
		 	* ==> sort out a few desktop PIIIs. */
			rdmsr(MSR_IA32_EBL_CR_POWERON, msr_lo, msr_hi);
			ddprintf(("Coppermine: MSR_IA32_EBL_CR_POWERON is 0x%x, 0x%x\n", msr_lo, msr_hi));
			msr_lo &= 0x00c0000;
			if (msr_lo != 0x0080000)
				return B_ERROR;
	
			/*
			 * If the processor is a mobile version,
			 * platform ID has bit 50 set
			 * it has SpeedStep technology if either
			 * bit 56 or 57 is set
			 */
			rdmsr(MSR_IA32_PLATFORM_ID, msr_lo, msr_hi);
			ddprintf(("Coppermine: MSR_IA32_PLATFORM ID is 0x%x, 0x%x\n", msr_lo, msr_hi));
			if ((msr_hi & (1<<18)) && (relaxed_check ? 1 : (msr_hi & (3<<24)))) {
				/*if (c->x86_mask == 0x01) {
					ddprintf(("early PIII version\n"));
					processorType = SPEEDSTEP_PROCESSOR_PIII_C_EARLY;
					return B_OK;
				} 
				else{
					processorType = SPEEDSTEP_PROCESSOR_PIII_C;
					return B_OK;
				}*/
				processorType = SPEEDSTEP_PROCESSOR_PIII_C;
				//return B_OK;
			}
			else{
				return B_ERROR;
			}
        	break;
        default:
        	return B_ERROR;
	}
	
	
	//all test succed
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
status_t init_driver (void)
{	
	int ix = 0;
	int cards_found = 0;
	
	ddprintf(("SPEEDSTEP init_driver\n"));
		
	//detect chipset
	if (get_module(B_PCI_MODULE_NAME, (module_info **)&pci) < 0)
	return ENOSYS;
	while ((*pci->get_nth_pci_info)(ix, &info) == B_OK) {
		//4-M
		if (info.vendor_id == PCI_VENDOR_ID_INTEL && info.device_id == PCI_DEVICE_ID_INTEL_82801DB_12) {
			cards_found++;
			//my_card_array[ix].info = info;
			ddprintf(("SPEEDSTEP card found 4-M\n"));
			break;
		}
		//3-M
		if (info.vendor_id == PCI_VENDOR_ID_INTEL && info.device_id == PCI_DEVICE_ID_INTEL_82801CA_12) {
			cards_found++;
			//my_card_array[ix].info = info;
			ddprintf(("SPEEDSTEP card found 3-M\n"));
			break;
		}
		//2-M
		if (info.vendor_id == PCI_VENDOR_ID_INTEL && info.device_id == PCI_DEVICE_ID_INTEL_82801BA_10) {
			
			//my_card_array[ix].info = info;
			/* speedstep.c causes lockups on Dell Inspirons 8000 and
			 * 8100 which use a pretty old revision of the 82815
			 * host brige. Abort on these systems.
		 	*/
		 	/*
			static struct pci_dev *hostbridge;
			u8 rev = 0;

			hostbridge  = pci_get_subsys(PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_82815_MC,
			      PCI_ANY_ID,
			      PCI_ANY_ID,
			      NULL);

			if (!hostbridge)
				return 2; // 2-M 

			pci_read_config_byte(hostbridge, PCI_REVISION_ID, &rev);
			if (rev < 5) {
				dprintk("hostbridge does not support speedstep\n");
				speedstep_chipset_dev = NULL;
				pci_dev_put(hostbridge);
				return 0;
			}

			pci_dev_put(hostbridge);
			return 2; 	//2-M 
			
			cards_found++;
			ddprintf(("SPEEDSTEP card found 2-M\n"));
			*/
			break;
		}
		
		ix++;
		//if (cards_found == MAX_CARDS) break;
	} 
	if(cards_found == 0){
		return ENODEV;
	}
	
	/* activate speedstep support */
	if (speedstep_activate(&info) != B_OK) {
		//pci_dev_put(speedstep_chipset_dev);
		return EINVAL;
	}
	
	if (speedstep_find_register(&info) != B_OK){
		return ENODEV;
	}
	
	speedstep_set_state (SPEEDSTEP_HIGH);
	cpuState.frequency = pentium3_get_frequency(processorType);
	cpuState.state = SPEEDSTEP_HIGH;
		
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
void uninit_driver (void)
{
}

	
// ----------------------------------------------------------------------------------------------
static status_t my_device_open (const char *name, uint32 flags, void** cookie)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	if(!buf){
		return B_ERROR;
	}
	((cpu_state*)buf)->state = cpuState.state;
	((cpu_state*)buf)->frequency = cpuState.frequency;
	return B_OK;
}

// ----------------------------------------------------------------------------------------------
static status_t my_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	*num_bytes = 0;
	return B_IO_ERROR;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	switch(op)
	{
		case SPEEDSTEP_HIGH:
			speedstep_set_state (SPEEDSTEP_HIGH);
			return B_OK;
		case SPEEDSTEP_LOW:
			speedstep_set_state (SPEEDSTEP_LOW);
			return B_OK;
		default:
			return B_BAD_VALUE;
	}
	return B_BAD_VALUE;
}


// ----------------------------------------------------------------------------------------------
static status_t my_device_close (void* cookie)
{
	return B_OK;
}

// ----------------------------------------------------------------------------------------------
static status_t my_device_free (void* cookie)
{
	return B_OK;
}


// ----------------------------------------------------------------------------------------------
static const char *my_device_name[] =
{
	"misc/speedstep_ich",
	NULL
};

// ----------------------------------------------------------------------------------------------
device_hooks my_device_hooks = {
	my_device_open, 		/* -> open entry point */
	my_device_close, 		/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

// ----------------------------------------------------------------------------------------------
const char** publish_devices()
{
	return my_device_name;
}

// ----------------------------------------------------------------------------------------------
device_hooks* find_device(const char* name)
{
	return &my_device_hooks;
}

