#include <stdlib.h>
#include <String.h>
#include <Box.h>

#include "prefwindow.h"

#include <Debug.h>

PrefWindow::PrefWindow(BRect rec, const char* title, BMessenger *target, const FreqConfig* config)
			 :BWindow(rec, title, B_TITLED_WINDOW,
			 	B_NOT_RESIZABLE | B_NOT_ZOOMABLE | B_AVOID_FRONT,
				B_ALL_WORKSPACES)
{
	fConfig = new FreqConfig();
	*fConfig = *config;
	
	fMessenger = target;
	
	fMainLayout = new Layout(Bounds(), "main layout", B_FOLLOW_ALL);
	fMainLayout->SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	
	fAutoButton = new BRadioButton(BRect(0,0,50,16), 
									"auto", "Automatic change the frequency.", new BMessage(CONTROL_CHANGE));	
	fHighButton = new BRadioButton(BRect(0,0,50,16), 
									"high", "Always high frequency.", new BMessage(CONTROL_CHANGE));
	fLowButton = new BRadioButton(BRect(0,0,50,16), 
									"low", "Always low frequency.", new BMessage(CONTROL_CHANGE));
	
	BString out;
	out << config->integrationTime / 1000000;
	fIntegrationTimeControl = new BTextControl(BRect(0,0,Bounds().Width(),10),
					"intergal","Integration Time [s]",out.String(),new BMessage(INTEGRATIONTIME_CHANGED));
	fIntegrationTimeControl->SetModificationMessage(new BMessage(INTEGRATIONTIME_CHANGED));
	out = "";
	out << config->threshold * 100;
	fTresholdControl = new BTextControl(BRect(0,0,Bounds().Width(),10),
					"threshold","Threshold [%]",out.String(),new BMessage(THRESHOLD_CHANGED));
	fTresholdControl->SetModificationMessage(new BMessage(THRESHOLD_CHANGED));
	
	fMainLayout->AddView(fAutoButton, 19);
	fMainLayout->AddView(fIntegrationTimeControl,19);
	fMainLayout->AddView(fTresholdControl,19);
	fMainLayout->AddSpacer(5);
	fMainLayout->AddView(fHighButton, 19);
	fMainLayout->AddView(fLowButton, 19);	
	
	AddChild(fMainLayout);
	
	BList controlList;
	controlList.AddItem(fIntegrationTimeControl);
	controlList.AddItem(fTresholdControl);
	DivideSame(&controlList);
	if(fConfig->dynamicFreq){
		fAutoButton->SetValue(B_CONTROL_ON);
	}
	if(fConfig->highState){
		fHighButton->SetValue(B_CONTROL_ON);
	}
	if(fConfig->lowState){
		fLowButton->SetValue(B_CONTROL_ON);
	}
}


PrefWindow::~PrefWindow()
{
	delete fMessenger;
	delete fConfig;
}


void	
PrefWindow::MessageReceived(BMessage *msg)
{
	bool somethingChanged = false;
	switch(msg->what) {
		case CONTROL_CHANGE:
			fConfig->dynamicFreq = fAutoButton->Value();
			fConfig->highState = fHighButton->Value();
			fConfig->lowState = fLowButton->Value();
			somethingChanged = true;
			break;		
		case INTEGRATIONTIME_CHANGED:
			fConfig->integrationTime = atoi(fIntegrationTimeControl->Text()) * 1000000;
			somethingChanged = true;
			break;
		case THRESHOLD_CHANGED:
			fConfig->threshold = atof(fTresholdControl->Text()) / 100;
			somethingChanged = true;
			break;
	}
	
	if(somethingChanged){
		BMessage msg(NEW_CONFIG);
		msg.AddPointer("config", fConfig);
		fMessenger->SendMessage(&msg);
	}
}


bool 
PrefWindow::QuitRequested( void )
{
	fMessenger->SendMessage(PREF_EXITED);
	return(TRUE);
}


void	
PrefWindow::SetTarget(BHandler *handler)
{
	delete fMessenger;
	fMessenger = new BMessenger(handler);
}