#include "speedstep.h"
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <String.h>
#include <interface/Deskbar.h>
#include <Alert.h>
#include <Path.h>
#include <stdlib.h>

#include <Debug.h>


FreqConfig::FreqConfig()
{
	dynamicFreq = true; 
	lowState = false;
	highState = false;
	integrationTime = 2000000;
	threshold = 0.95;
}


SpeedStepView::SpeedStepView(BRect frame, const char *name, const char *text)
							:BStringView(frame, name, text)
{
	BPath path;
	find_directory(B_USER_SETTINGS_DIRECTORY, &path);
	path.Append("SpeedStepp_ICH_settings");
	fConfigFile = new BFile(path.Path(), B_READ_WRITE | B_CREATE_FILE);
	
	LoadConfig();
	BuildView();
}


SpeedStepView::SpeedStepView(BMessage *message)
	:BStringView(message)
{
	BPath path;
	find_directory(B_USER_SETTINGS_DIRECTORY, &path);
	path.Append("SpeedStepp_ICH_settings");
	fConfigFile = new BFile(path.Path(), B_READ_WRITE | B_CREATE_FILE);
	
	LoadConfig();
	BuildView();
}


SpeedStepView::~SpeedStepView()
{
	if(fDriverHandler){
		close(fDriverHandler);
	}
	delete fConfigFile;
	if(fMessageRunner){
		delete fMessageRunner;
	}
}


void	
SpeedStepView::MessageReceived(BMessage *msg)
{
	FreqConfig *config = NULL;
	switch(msg->what) {
		case NEW_CONFIG:
			msg->FindPointer("config",&((void*)config));
			NewConfig(config);	
			break;	
		case PREF_EXITED:
			fPrefWindow = NULL;
			break;
		case REPLICANT_PULSE:
			PrepPulse();
			break;
		case PREF_ITEM_INVOKED:
			if(fPrefWindow){
				fPrefWindow->Activate(true);
				break;
			}
			fPrefWindow = new PrefWindow(BRect(200,100,400,205), "Speedstep Preferences", 
											new BMessenger(this), &fConfig);
			//fPrefWindow->SetTarget(this);
			fPrefWindow->Show();
			break;
		case QUIT_ITEM_INVOKED:
			Remove();
			break;
		default:
			BView::MessageReceived(msg);
	}
	
}


void
SpeedStepView::MouseDown(BPoint point)
{
	ShowPoUpMenu(&point);
}


void	
SpeedStepView::AttachedToWindow()
{
	fItemPref->SetTarget(this);
	fItemQuit->SetTarget(this);
	
	system_info sysInfo;
	get_system_info(&sysInfo);
	fPrevActive = sysInfo.cpu_infos[0].active_time;
	fPrevTime = system_time();
	fMessageRunner = new BMessageRunner(this, new BMessage(REPLICANT_PULSE),
										fConfig.integrationTime, -1);
}


SpeedStepView*
SpeedStepView::Instantiate(BMessage *data)
{
	PRINT(("halloo\n"));
	if (!validate_instantiation(data, "SpeedStepView")){
		PRINT(("halloo null\n"));
		return NULL;
	}
	return new SpeedStepView(data);
}


status_t 
SpeedStepView::Archive(BMessage *data, bool deep = true) const
{
	BStringView::Archive(data, deep);
	data->AddString("add_on", APP_SIGNATURE);
	data->AddString("class", "SpeedStepView");
	return B_OK;
}


void	
SpeedStepView::PrepPulse(void)
{
	if(!fConfig.dynamicFreq){
		return;
	}
	
	system_info sysInfo;
	get_system_info(&sysInfo);
	bigtime_t now = system_time();
	double usage = (double)(sysInfo.cpu_infos[0].active_time - fPrevActive ) / (now - fPrevTime);
	PRINT(("hello usage %f %f %f\n", usage,sysInfo.cpu_infos[0].active_time, fPrevActive));
	if(usage > fConfig.threshold){
		if(fCpuState.state == SPEEDSTEP_LOW)
		{
			PRINT(("set low speed\n"));
			SetState(SPEEDSTEP_HIGH);
			UpdateView();
		}
	}
	else if(fCpuState.state == SPEEDSTEP_HIGH)
	{
		PRINT(("set high speed\n"));
		SetState(SPEEDSTEP_LOW);
		UpdateView();
	}
	
	PRINT(("frec %i\n", fCpuState.frequency));
	fPrevActive = sysInfo.cpu_infos[0].active_time;
	fPrevTime = now;
	
	
}


status_t	
SpeedStepView::SetState(int32 state)
{
	if(!fDriverHandler){
		return B_ERROR;
	}
	status_t ret;
	ret = ioctl(fDriverHandler, state, NULL, 0);
	read(fDriverHandler, &fCpuState, sizeof(fCpuState));
	return ret;
}


void
SpeedStepView::ShowPoUpMenu(BPoint *point)
{
	ConvertToScreen(point);
	fMainMenu->Go(*point, true);
}


void SpeedStepView::Remove() {
	// Remove ourselves from the deskbar by name
	BDeskbar *deskbar = new BDeskbar();
	status_t err = deskbar->RemoveItem("SpeedStepView");
	if (err != B_OK) {
		char temp[255];
		sprintf(temp, "Remove(): %s", strerror(err));
		BAlert *alert = new BAlert("Info", temp, "OK");
		alert->Go(NULL);
	}
	delete deskbar;
}


void	
SpeedStepView::UpdateView()
{
	if(fCpuState.state == SPEEDSTEP_LOW){
		SetLowColor(kLowStateColor);
		SetViewColor(kLowStateColor);
	}
	else{
		SetLowColor(kHighStateColor);
		SetViewColor(kHighStateColor);
	}
	BString freq;
	freq << fCpuState.frequency / 1000;
	freq+= " Mhz";
	SetText(freq.String());
	ResizeTo(StringWidth(freq.String()), 16);
	
}


void	
SpeedStepView::BuildView()
{
	fPrefWindow = NULL;
	fMessageRunner = NULL;	
	//mainview
	SetText("now driver found");
	
	//popupmenu
	fMainMenu = new BPopUpMenu("main menu");
	fItemPref = new BMenuItem("Preferences",new BMessage(PREF_ITEM_INVOKED));
	fItemQuit = new BMenuItem("Quit",new BMessage(QUIT_ITEM_INVOKED));
	fMainMenu->AddItem(fItemPref);
	fMainMenu->AddItem(fItemQuit);
	
	//read from driver
	fDriverHandler = 0;
	fDriverHandler = open("/dev/misc/speedstep_ich", O_RDWR | O_NONBLOCK);
	if(fDriverHandler) {
		//status_t ret;
		//ret = ioctl(fDriverHandler, SPEEDSTEP_LOW, NULL, 0);
	
		read(fDriverHandler, &fCpuState, sizeof(fCpuState));
		UpdateView();
	}
}


void	
SpeedStepView::NewConfig(const FreqConfig* config)
{
	if(!config){
		return;
	}
	fConfig = *config;
	
	if(fConfig.lowState == true){
		SetState(SPEEDSTEP_LOW);
		UpdateView();
	}
	
	if(fConfig.highState == true){
		SetState(SPEEDSTEP_HIGH);
		UpdateView();
	}
	
	SaveConfig();
}


void	
SpeedStepView::LoadConfig()
{
	status_t err = fConfigFile->ReadAttr("autoControl", B_BOOL_TYPE, 0, &(fConfig.dynamicFreq), 1);
	err = fConfigFile->ReadAttr("lowState", B_BOOL_TYPE, 0, &(fConfig.lowState), 1);
	err = fConfigFile->ReadAttr("highState", B_BOOL_TYPE, 0, &(fConfig.highState), 1);
	err = fConfigFile->ReadAttr("integrationTime", B_INT64_TYPE, 0, &(fConfig.integrationTime), 8);
	err = fConfigFile->ReadAttr("threshold", B_FLOAT_TYPE, 0, &(fConfig.threshold), 4);
}
		
				
void	
SpeedStepView::SaveConfig()
{
	status_t err = fConfigFile->WriteAttr("autoControl", B_BOOL_TYPE, 0, &(fConfig.dynamicFreq), 1);
	err = fConfigFile->WriteAttr("lowState", B_BOOL_TYPE, 0, &(fConfig.lowState), 1);
	err = fConfigFile->WriteAttr("highState", B_BOOL_TYPE, 0, &(fConfig.highState), 1);
	err = fConfigFile->WriteAttr("integrationTime", B_INT64_TYPE, 0, &(fConfig.integrationTime), 8);
	err = fConfigFile->WriteAttr("threshold", B_FLOAT_TYPE, 0, &(fConfig.threshold), 4);
	
}

				
SpeedStepApp::SpeedStepApp(const char *signature):BApplication(signature)
{
	BDeskbar *deskbar = new BDeskbar();
	if (deskbar->HasItem("SpeedStepView")) {
		//deskbar->RemoveItem("SpeedStepView");
		delete deskbar;
		return;
	}
	SpeedStepView	*speedStepView = new SpeedStepView(BRect(0,0,40,16), "SpeedStepView" , "Mhz View");
	//BStringView *stringView = new BStringView(BRect(0,0,10,16),"test","test");
	status_t err = deskbar->AddItem(speedStepView);
	delete speedStepView;
	delete deskbar;
	if (err != B_OK) {
		BAlert *alert = new BAlert(NULL, strerror(err), "OK");
		alert->Go(NULL);
		return;
	}
}


SpeedStepApp::~SpeedStepApp() {
	// Load the replicant after we save our preferences so they don't
	// get overwritten by DeskbarPulseView's instance
	
}




