#ifndef SPEEDSTEP_H
#define SPEEDSTEP_H

#include <Application.h>
#include <Window.h>
#include <View.h>
#include <StringView.h>
#include <MessageRunner.h>
#include <File.h>
#include <FindDirectory.h>

#include "prefwindow.h"
#include "../Driver/SpeedStep.h"
#include "common.h"

const uint REPLICANT_PULSE	= '&rep';
const uint QUIT_ITEM_INVOKED = '&qui';
const uint PREF_ITEM_INVOKED = '&pri';

const rgb_color kLowStateColor={0,0, 255,0};
const rgb_color kHighStateColor={255,0,0,0};
const rgb_color kBlackColor={0,0,0,0};


class SpeedStepView : public BStringView
{
	public:
						SpeedStepView(BRect frame, const char *name, const char *text);
						SpeedStepView(BMessage *message);
						~SpeedStepView();
		virtual	void	MessageReceived(BMessage *msg);
		virtual void	MouseDown(BPoint point);
		virtual void	AttachedToWindow();
	static SpeedStepView*	Instantiate(BMessage *data);
		virtual	status_t Archive(BMessage *data, bool deep = true) const;
				void	PrepPulse();
				void 	ShowPoUpMenu(BPoint *point);
				void	Remove();
				void	UpdateView();
				void	SetTarget(BHandler *handler);
			status_t	SetState(int32 state);
	private:
				void	BuildView();
				void	NewConfig(const FreqConfig* config);
				void	LoadConfig();
				void	SaveConfig();
		
		PrefWindow*		fPrefWindow;
		BPopUpMenu*		fMainMenu;
		BMenuItem*		fItemQuit;
		BMenuItem*		fItemPref;
		BMessageRunner*	fMessageRunner;
		cpu_state		fCpuState;
		FreqConfig		fConfig;
		bigtime_t 		fPrevActive;
		bigtime_t 		fPrevTime;
		int32 			fDriverHandler;
		BFile			*fConfigFile;
};


class SpeedStepApp  : public BApplication {
	public:
						SpeedStepApp(const char *signature);
						~SpeedStepApp();
		
};


#endif